package kc.mega.learning;

import kc.mega.wave.WaveWithFeatures;

import java.util.Arrays;
import java.util.List;

import ags.utils.KdTree.Entry;

public class KNNAimModel extends AimModel {
  private final WaveKNN<double[]> knn;

  public KNNAimModel(String name, WaveKNN<double[]> knn) {
    super(name);
    this.knn = knn;
  }

  @Override
  public boolean hasData() {
    return knn.isEmpty();
  }

  @Override
  public void train(WaveWithFeatures w) {
    knn.addPoint(w, w.hitGFRange());
  }

  @Override
  public double getAimAngle(WaveWithFeatures w) {
    return getAimAngle(w, knn.getNumNeighbors());
  }

  @Override
  public double getAimAngleFast(WaveWithFeatures w) {
    return getAimAngle(w, knn.getNumNeighbors() / 2);
  }

  private double getAimAngle(WaveWithFeatures w, int numNeighbors) {
    List<Entry<double[]>> neighbors = knn.getNeighbors(w, numNeighbors);
    if (neighbors.size() == 0) {
      return w.absoluteBearing;
    }

    double[] weights = knn.getWeights(neighbors);
    HitEdge[] edges = new HitEdge[2 * neighbors.size()];
    for (int i = 0; i < neighbors.size(); i++) {
      double[] gfRange = neighbors.get(i).value;
      edges[2 * i] = new HitEdge(gfRange[0], weights[i]);
      edges[1 + 2 * i] = new HitEdge(gfRange[1], -weights[i]);
    }
    Arrays.sort(edges);
    double height = 0;
    double bestHeight = 0;
    int bestIndex = 0;
    for (int i = 0; i < edges.length; i++) {
      height += edges[i].weight;
      if (height > bestHeight) {
        bestHeight = height;
        bestIndex = i;
      }
    }

    double GF = (edges[bestIndex].GF + edges[bestIndex + 1].GF) / 2;
    return w.absoluteBearing + w.getOffset(GF);
  }

  public static class HitEdge implements Comparable<HitEdge> {
    public double GF;
    public double weight;

    public HitEdge(double GF, double weight) {
      this.GF = GF;
      this.weight = weight;
    }

    @Override
    public int compareTo(HitEdge gfScore){
      return (int)Math.signum(GF - gfScore.GF);
    }
  }
}
