package kc.mega.wave;

import kc.mega.game.BattleField;
import kc.mega.game.BotState;
import kc.mega.learning.APMFlattener;
import kc.mega.utils.Geom;

import java.awt.geom.Point2D;
import java.util.List;

import jk.math.FastTrig;

public class WaveWithFeatures extends Wave {
  public final BotState fireState;
  public final BotState myState;
  public final BotState enemyState;
  public final BotState prevEnemyState;

  public final double prevAbsoluteBearing;
  public final int moveDirection;
  public final double distance;
  public final double accel;
  public final double velocity;
  public final double avgVelocity;
  public final double relativeHeading;
  public final int vChangeTimer;
  public final int dirChangeTimer;
  public final int decelTimer;
  public final double distanceLast10;
  public final double distanceLast20;
  public final double stickWallAhead;
  public final double stickWallReverse;
  public final double maeWallAhead;
  public final double maeWallReverse;
  public final double orbitalWallAhead;
  public final double orbitalWallReverse;
  public final int shotsFired;
  public final double virtuality;
  public final double currentGF;
  public final String pattern;

  public static class Builder extends Wave.Builder {
    public WaveManager.BotWaveManager<?> waveManager;
    public Builder waveManager(WaveManager.BotWaveManager<?> val) {waveManager = val; return this;}
    @Override
    public WaveWithFeatures build() {return new WaveWithFeatures(this);}
  }

  public WaveWithFeatures(Builder builder) {
    super(builder);

    fireState = builder.myHistory.get(0);
    Point2D.Double myNextLocation = fireState.location;
    // start at 1 because aiming was done the tick before the wave is created
    List<BotState> myHistory = builder.myHistory.subList(1, builder.myHistory.size());
    List<BotState> enemyHistory = builder.enemyHistory.subList(1, builder.enemyHistory.size());

    myState = myHistory.get(0);
    enemyState = enemyHistory.get(0);
    prevEnemyState = enemyHistory.get(1);
    Point2D.Double myLocation = myState.location;
    Point2D.Double enemyLocation = enemyState.location;

    double gunHeat = Math.max(myState.gunHeat, 0);
    int ticksSinceShot = hasBullet ? 0 : (int)(myState.gameTime - myState.lastFireTime);
    double coolingRate = BattleField.INSTANCE.getGunCoolingRate();
    virtuality = hasBullet ? 0 : Math.min((int)(gunHeat / coolingRate), 1 + ticksSinceShot);
    didHit = false;
    shotsFired = myState.shotsFired;
    currentGF = builder.waveManager == null ? 0 : builder.waveManager.getCurrentGF(enemyState);

    prevAbsoluteBearing = Geom.absoluteBearingPrecise(myLocation, enemyLocation);
    distance = myLocation.distance(enemyLocation);
    velocity = enemyState.velocity;
    relativeHeading = enemyState.heading - prevAbsoluteBearing;
    accel = Math.abs(enemyState.velocity - prevEnemyState.velocity) * (
        Math.abs(enemyState.velocity) < Math.abs(prevEnemyState.velocity) ? -1 : 1);

    int vChangeTimerr = enemyHistory.size();
    int dirChangeTimerr = enemyHistory.size();
    int decelTimerr = enemyHistory.size();
    int moveDirectionn = (int)Math.signum(enemyState.velocity);
    double totalVelocity = 0;
    double totalWeight = 0;
    double weight = 1;
    String patternn = "";
    for (int i = 0; i < Math.min(81, enemyHistory.size()); i++) {
      BotState enemyState = enemyHistory.get(i);
      double vel = enemyState.velocity;
      totalVelocity += weight * vel;
      totalWeight += weight;
      weight *= 0.95;
      if (i > 0 && Math.abs(vel - enemyHistory.get(i - 1).velocity) > 0.01 &&
          i - 1 < vChangeTimerr) {
        vChangeTimerr = i - 1;
      }
      if (vel != 0) {
        if (moveDirectionn == 0) {
          moveDirectionn = (int)Math.signum(vel);
        } else if((int)Math.signum(vel) != moveDirectionn && i - 1 < dirChangeTimerr) {
          dirChangeTimerr = i - 1;
        }
      }
      if (moveDirectionn != 0 && i > 0 &&
          enemyHistory.get(i - 1).velocity * moveDirectionn < vel * moveDirectionn &&
          i - 1 < decelTimerr) {
        decelTimerr = i - 1;
      }
      if (i < myHistory.size()) {
        patternn += APMFlattener.getPMSymbol(myHistory.get(i), enemyState);
      }
    }
    if (moveDirectionn == 0) {
      moveDirectionn = 1;
    }
    moveDirection = moveDirectionn;
    vChangeTimer = vChangeTimerr;
    dirChangeTimer = dirChangeTimerr;
    decelTimer = decelTimerr;
    pattern = patternn;
    avgVelocity = totalVelocity / (totalWeight == 0 ? 1 : totalWeight);
    distanceLast10 = enemyLocation.distance(
        enemyHistory.get(Math.min(12, enemyHistory.size() - 1)).location);
    distanceLast20 = enemyLocation.distance(
        enemyHistory.get(Math.min(20, enemyHistory.size() - 1)).location);

    double maxEscapeAngle = Geom.simpleMaxEscapeAngle(speed);
    orbitalWallAhead = BattleField.INSTANCE.orbitalWallDistance(
        myState, enemyState, orbitDirection, maxEscapeAngle, true);
    orbitalWallReverse = BattleField.INSTANCE.orbitalWallDistance(
        myState, enemyState, orbitDirection, maxEscapeAngle, false);
    double[] inFieldMaes = Geom.inFieldMaxEscapeAngle(myNextLocation, enemyLocation, speed);
    maeWallAhead = (orbitDirection < 0 ? inFieldMaes[0] : inFieldMaes[1]) / maxEscapeAngle;
    maeWallReverse = (orbitDirection > 0 ? inFieldMaes[0] : inFieldMaes[1]) / maxEscapeAngle;
    double targetHeading = absoluteBearing + orbitDirection * moveDirection * Math.PI / 2;
    double smoothedForward = BattleField.INSTANCE.walkingStickSmooth(
        enemyLocation, targetHeading, moveDirectionn, orbitDirection, 160, 18);
    double smoothedBackward = BattleField.INSTANCE.walkingStickSmooth(
        enemyLocation, targetHeading, -moveDirectionn, -orbitDirection, 160, 18);
    stickWallAhead = Math.abs(FastTrig.normalRelativeAngle(smoothedForward - targetHeading));
    stickWallReverse = Math.abs(FastTrig.normalRelativeAngle(smoothedBackward - targetHeading));
  }

  public double[] toArray() {
    return new double[] {
        absoluteBearing,
        orbitDirection,
        moveDirection,
        distance,
        power,
        accel,
        Math.signum(accel),
        relativeHeading,
        velocity,
        vChangeTimer,
        dirChangeTimer,
        decelTimer,
        currentGF,
        hasBullet ? 1 : 0,
        didHit ? 1 : 0,
        didCollide ? 1 : 0,
        shotsFired,
        fireTime,
        distanceLast10,
        distanceLast20,
        orbitalWallAhead,
        orbitalWallReverse,
        maeWallAhead,
        maeWallReverse,
        stickWallAhead,
        stickWallReverse,
        virtuality,
        maePrecise[0],
        maePrecise[1],
        hitOffset,
        aggHitOffsetRange[0],
        aggHitOffsetRange[1]
    };
  }
}
