/*
FoilistMicro v2.0 by Sheldor.  09/17/2023  749 bytes
multimode movement and basic dynamic clustering targeting
v2.0 -- major changes to targeting and movement

Foil is one of the three forms of modern sport fencing,
along with epee and sabre.  http://en.wikipedia.org/wiki/Foil_%28fencing%29
 
Credits:
A general thanks to all open source bot authors and contributors to the RoboWiki.  Without them, Robocode would not be nearly as advanced as it currently is.
Bots which I'd like to specifically credit include, but are not limited to:
	targeting: jk.mini.CunobelinDC, Falcon, pez.micro.Aristocles, kc.micro.Thorn, jk.micro.Connavar, nz.jdc.HedgehogGF, pez.mini.Pugilist, sheldor.micro.Epeeist
	movement : jk.micro.Cotillion, jk.micro.Toorkild, kc.micro.Thorn, nz.jdc.HedgehogGF, wiki.nano.RaikoNano

FoilistMicro is released under the terms of the RoboWiki Public Code License (RWPCL) - Version 1.1.
see license here:  http://robowiki.net/wiki/RWPCL
*/

package sheldor.micro;

import robocode.*;
import robocode.util.Utils;
import java.awt.geom.*;
import java.util.ArrayList;
import java.util.*;

public class FoilistMicro extends AdvancedRobot
{
	//constants
	static final int    GUESS_FACTORS = 29;
	static final int    MIDDLE_FACTOR = (GUESS_FACTORS - 1) / 2;
	static final double MAXIMUM_ESCAPE_ANGLE = 8.0 / 14;
	static final double FACTOR_ANGLE = MAXIMUM_ESCAPE_ANGLE / MIDDLE_FACTOR;
	
	public static final char FW1 = 4;
	public static final char FW2 = 9;
	public static final char FW3 = 16;
	public static final char RW1 = (char)-9;
	
	public static final char VELOCITY_VALUE = 768;
	public static final int  WALL_VALUE = 6;
	
	//global variables
	static double direction = 1;
	static double enemyBulletSpeed;
	static double enemyDirection;
	static double enemyEnergy;
	static double hits;
	static int    movementMode;
	
	static int ticksSinceVelocityChange;
	static int previousEnemyVelocity;
	
	static double enemyHeading;
	static double enemyVelocity;
	
	static ArrayList<Wave> waves = new ArrayList();
	
	public void onStatus(StatusEvent e)
	{		
		//turn the radar every tick
		//Putting the code here instead of in a while(true) loop in the run() method saves one byte.
		//I believe Wompi discovered this.
    	setTurnRadarRightRadians(1);
		
		//set the radar and gun to turn independently
		//Putting these here with the radar code lets me get rid of the run() method, saving one byte.  credit to Cotillion
		setAdjustRadarForGunTurn(true);
		setAdjustGunForRobotTurn(true);
	}
	
	public void onScannedRobot(ScannedRobotEvent e)
	{
		//local variables
		int    i;		
		double offset;
		double enemyDistance;
		double absoluteBearing;
		double bulletPower;		
		double lateralVelocity;	
		double localEnemyDirection;	
		double theta;
		int    integer;
		
		Wave   wave = new Wave();
		
		if (setFireBullet((offset = 
			(2 + (integer = ((i = 112) / (int)(wave.enemyDistance = enemyDistance = e.getDistance())))))
			- BULLET_POWER_TABLE.charAt((int)getEnergy())) != null){
			addCustomEvent(wave);
		}
				
		/*********************************************
		 *---------------MOVEMENT CODE---------------*
		 *********************************************/				
		
		//wall smoothing based on HedgehogGF's
		while(fieldContains(theta = (wave.absoluteBearing = absoluteBearing = 
			(e.getBearingRadians() + getHeadingRadians())) + direction * (offset -= 0.02), 160) > 0);
		setTurnRightRadians(Math.tan(theta -= getHeadingRadians()));
			
		//stop and go movement, now fixed-length for code size
		//move when the enemy fires, or when the robot is moving randomly, or when the enemy is very close
		if (enemyEnergy - (enemyEnergy = e.getEnergy()) > movementMode - integer)
		{
			//credit to HedgehogGF for the copySign trick
			setAhead(Math.copySign(48, Math.cos(theta)));
		}
				
		//random movement from Toorkild
		//don't move randomly if the enemy is ramming
		//reverse direction if the bot gets too close to a wall
		if (Math.random() + integer < (-0.6 * Math.sqrt(enemyBulletSpeed / enemyDistance) + 0.04)
			* movementMode	|| offset < Math.PI/3.5)
		{
			direction = -direction;
		}
		
		/********************************************
		 *--------------TARGETING CODE--------------*
		 ********************************************/
		
		//determine enemy wall proximity
		//inspired by Pugilist and HedgehogGF
		int wallScore = 0;	
		do
		{
			wallScore += fieldContains(absoluteBearing + (
			
					//determine the enemy's lateral movement direction
					//use a simple rolling average to store the previous lateral direction if enemy lateral velocity == 0
					//credit to HedgehogGF
					(wave.enemyDirection = localEnemyDirection = (enemyDirection = Math.signum(0.00000000000001 +
					 ((enemyVelocity = e.getVelocity()) * (Math.sin((enemyHeading = e.getHeadingRadians()) - absoluteBearing)))
					 + (enemyDirection / 100))) * FACTOR_ANGLE)

				 * ((short)WALL_TABLE.charAt(i))), enemyDistance);
		}
		while(--i > 0);
		
		//calculate deceleration and whether or not to reset ticks since velocity change
		int deceleration;
		if ((deceleration = Integer.signum(previousEnemyVelocity - (previousEnemyVelocity = (int)Math.abs(enemyVelocity)))) != 0)
		{
			ticksSinceVelocityChange = deceleration;
		}
		
		//determine current situation
		double localScore = wave.score =
			VELOCITY_HISTORY_TABLE.charAt(ticksSinceVelocityChange += 3) +
			(wallScore * WALL_VALUE) +
			((previousEnemyVelocity / 3) << 5) +
			enemyDistance / 60;
			
		//calculate angle scores and select aim angle
		int j = -MIDDLE_FACTOR;
		offset = 0;
		do{
			i = 0;
			double angleScore = 0;
			try{
				while(true)
				{
					if (Math.abs(j - (wave = waves.get(i++)).result) < 1){
						angleScore += i / (1 + Math.abs(wave.score - localScore));
					}
				}
			}catch (Exception ex){};	
		
			if (angleScore >= offset){
				offset = angleScore;	
				setTurnGunRightRadians((0.007 * Math.random()) + Utils.normalRelativeAngle(absoluteBearing
					 - getGunHeadingRadians()	+ (localEnemyDirection * j)));
			}
		}while (++j < MIDDLE_FACTOR);
		
		//radar lock
		setTurnRadarRightRadians(2 * Utils.normalRelativeAngle(absoluteBearing - getRadarHeadingRadians()));	
	}
	
	public void onBulletHit(BulletHitEvent e)
	{
		//adjust the enemy energy variable when the bot hits the enemy
		//this makes a big difference against linear targeting
		enemyEnergy -= 10;
	}
	
	public void onHitByBullet(HitByBulletEvent e)
	{
		//adjust the enemy energy variable when the bot gets hit
		//store the velocity of the enemy's bullet for the random movement
		enemyEnergy += 20 - (enemyBulletSpeed = e.getVelocity());
		
		//mode selection inspired by Toorkild
		//if the bot takes an unacceptable amount of damage relative to the number of rounds
		//that have passed while in stop and go mode, switch to random movement
		if ((hits += 0.25) > getRoundNum() + 1)
		{
			movementMode = -1;
		}
    }
	
	//This method returns 1 if a point projected from the bot's location by the 
	//"heading" and "distance" parameters is outside of the battlefield, and 0 if it is not.
	//credit to HedgehogGF
	private int fieldContains(double heading, double distance)
	{
		return Integer.signum(new Rectangle2D.Double(18, 18, 764, 564).outcode(getX() + distance * Math.sin(heading), getY() + distance * Math.cos(heading)));
	}
	
	static class Wave extends Condition
	{
		//global variables
		double absoluteBearing;
		double enemyDirection;
		double bearingOffset;
		double enemyDistance;
		double waveDistanceTraveled;
		double score;
		double result;			
		
		public boolean test()
		{
			double relativeHeading;
			bearingOffset += ((enemyVelocity * Math.sin(relativeHeading = (enemyHeading - absoluteBearing))) / (enemyDistance += (enemyVelocity * Math.cos(relativeHeading))));
			
			//check if the wave has passed the enemy's current location
		    if ((waveDistanceTraveled += 14) > enemyDistance)
			{
				result = bearingOffset / enemyDirection;
				
				waves.add(this);
				waveDistanceTraveled = Double.NEGATIVE_INFINITY;
			}
			return false;
		}
	}
	
	//table for wall tests
	static final String WALL_TABLE = ""
	+ (char)MIDDLE_FACTOR
	
	+ FW1 + FW1 + FW1 + FW1 + FW1
	+ FW1 + FW1 + FW1 + FW1 + FW1
	+ FW1 + FW1 + FW1 + FW1 + FW1
	+ FW1 + FW1 + FW1 + FW1 + FW1
	+ FW1 + FW1 + FW1 + FW1 + FW1
	+ FW1 + FW1 + FW1 + FW1 + FW1
	+ FW1 + FW1
	
	+ FW2 + FW2 + FW2 + FW2 + FW2
	+ FW2 + FW2 + FW2 + FW2 + FW2
	+ FW2 + FW2 + FW2 + FW2 + FW2
	+ FW2 + FW2 + FW2 + FW2 + FW2
	+ FW2 + FW2 + FW2 + FW2 + FW2
	+ FW2 + FW2 + FW2 + FW2 + FW2
	+ FW2 + FW2
	
	+ FW3 + FW3 + FW3 + FW3 + FW3
	+ FW3 + FW3 + FW3 + FW3 + FW3
	+ FW3 + FW3 + FW3 + FW3 + FW3
	+ FW3 + FW3 + FW3 + FW3 + FW3
	+ FW3 + FW3 + FW3 + FW3 + FW3
	+ FW3 + FW3 + FW3 + FW3 + FW3
	+ FW3 + FW3
	
	+ RW1  + RW1  + RW1  + RW1
	+ RW1  + RW1  + RW1  + RW1
	+ RW1  + RW1  + RW1  + RW1
	+ RW1  + RW1  + RW1  + RW1
	;
		
	//tables for energy management
	static final String BP0 = ""
	+ (char)0 + (char)0 + (char)0 + (char)0 
	+ (char)0 + (char)0 + (char)0 + (char)0
	+ (char)0 + (char)0 + (char)0 + (char)0 
	+ (char)0 + (char)0 + (char)0 + (char)0
	+ (char)0 + (char)0 + (char)0 + (char)0 
	+ (char)0 + (char)0 + (char)0 + (char)0
	+ (char)0 + (char)0 + (char)0 + (char)0 
	+ (char)0 + (char)0 + (char)0 + (char)0;
	
	static final String	BULLET_POWER_TABLE = ""
	+ (char)2 + (char)2 + (char)2 + (char)2
	+ (char)2 + (char)1 + (char)1 + (char)1 
	+ (char)1 + (char)1 + (char)1 + (char)1
	+ (char)1 + (char)1 + (char)1 + (char)1
	+ (char)1 + (char)1 + (char)1 + (char)1
	+ BP0 + BP0 + BP0 + BP0 + BP0 + BP0 + BP0
	+ BP0 + BP0 + BP0 + BP0 + BP0 + BP0 + BP0
	+ BP0 + BP0 + BP0 + BP0 + BP0 + BP0 + BP0
	+ BP0 + BP0 + BP0 + BP0 + BP0 + BP0 + BP0
	+ BP0 + BP0 + BP0 + BP0 + BP0 + BP0 + BP0;	
	
	//velocity history tables
	//substrings for deceleration information
	static final String D0 = ""
	+ (char)4*VELOCITY_VALUE + (char)4*VELOCITY_VALUE + (char)4*VELOCITY_VALUE;
	
	static final String D1 = ""
	+ (char)3*VELOCITY_VALUE + (char)3*VELOCITY_VALUE + (char)3*VELOCITY_VALUE;
	
	static final String D2 = ""
	+ (char)2*VELOCITY_VALUE + (char)2*VELOCITY_VALUE + (char)2*VELOCITY_VALUE;
	
	static final String D3 = ""
	+ (char)1*VELOCITY_VALUE + (char)1*VELOCITY_VALUE + (char)1*VELOCITY_VALUE;
	
	static final String D4 = ""
	+ (char)0*VELOCITY_VALUE + (char)0*VELOCITY_VALUE + (char)0*VELOCITY_VALUE;
	
	static final String D4_ = ""
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4 
		+ D4 + D4 + D4 + D4 + D4 + D4 + D4;
	
	static final String VELOCITY_HISTORY_TABLE = ""	
		+ (char)0 + (char)0 + (char)5*VELOCITY_VALUE
		+ (char)0 + (char)6*VELOCITY_VALUE + (char)0
		
		+ D0 + D0 + D0 + D0 + D0
		
		+ D1 + D1 + D1 + D1 + D1
		+ D1 + D1 + D1 + D1 + D1
		
		+ D2 + D2 + D2 + D2 + D2
		+ D2 + D2 + D2 + D2 + D2
		+ D2 + D2 + D2 + D2 + D2
		+ D2 + D2 + D2 + D2 + D2
		
		+ D3 + D3 + D3 + D3 + D3
		+ D3 + D3 + D3 + D3 + D3
		+ D3 + D3 + D3 + D3 + D3
		+ D3 + D3 + D3 + D3 + D3
		+ D3 + D3 + D3 + D3 + D3
		
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_
		+ D4_ + D4_ + D4_ + D4_ + D4_ + D4_ + D4_;	
}